// TODO: Put public facing types in this file.

// import 'dart:async';
// import 'dart:io';

import 'package:kenken/kenken.dart';

/// Checks if you are awesome. Spoiler: you are.
class Kenken {
  Kenken(this.size)
      : cells = List.generate(size, (_) => List.generate(size, (_) => Cell()));

  /// The size of kenken. F.e. size=3 means that kenken has 3 rows and 3 columns.
  int size;

  Iterable<int> get rowOrColNumbers sync* {
    var i = 1;
    while (i <= size) {
      yield i++;
    }
  }

  /// Store for cell values, implemented as `List`(rows) of `List`(columns) of `int`(values).
  var cells = <List<Cell>>[];

  /// Custom the kenken rules
  var customRules = <Rule>[];

  /// Returns all kenken rules;
  List<Rule> get rules => generalRules.cast<Rule>() + customRules.cast<Rule>();

  /// Adds a one custom rule to kenken.
  /// Use this method to manually (by your own program logic) add custom rules to kenken.
  /// See also `notCoveredByCustomRulesCells` method.
  /// Sample:
  /// ```
  /// kenken.addRule([RC(1, 2), RC(1, 3)], Operation(sum, label: '+'), 5)
  /// ```
  void addRule(
      List<RC> cellsCoordinates, Operation operation, validOperationResult) {
    if (cellsCoordinates.length == 1) {
      // when single cell in rule:
      operation.showLabel = false;
      var r = cellsCoordinates[0].row;
      var c = cellsCoordinates[0].col;
      setValue(r, c, validOperationResult);
    }
    var rule = Rule(cellsCoordinates, operation, validOperationResult);
    customRules.add(rule);
    cellOf(rule.coordinatesToDisplayTheRule).doesDisplayRule = true;
    for (var rc in rule.cellsCoordinates) {
      cellOf(rc).itRule = rule;
      cellOf(rc).isTopBorderBold = !rule.hasRC(RC(rc.row - 1, rc.col));
      cellOf(rc).isBottomBorderBold = !rule.hasRC(RC(rc.row + 1, rc.col));
      cellOf(rc).isLeftBorderBold = !rule.hasRC(RC(rc.row, rc.col - 1));
      cellOf(rc).isRightBorderBold = !rule.hasRC(RC(rc.row, rc.col + 1));
    }
  }

  /// Generates common (general) kenken rules.
  /// These rules are general rules for all kenkens.
  /// They require from kenken unique numbers in columns and rows.
  /// Initially, kenken always contains general rules.
  /// The programmer does not have to worry about adding general rules to kenken.
  List<Rule> get generalRules {
    var rules = List.generate(size, (i) {
      var rowNumber = i + 1;
      var rowCoordinates = coordinatesOfRow(rowNumber);
      return Rule(
          rowCoordinates,
          Operation(isUniq, label: 'uniqueness in row', showLabel: false),
          true);
    });
    var colRules = List.generate(size, (i) {
      var colNumber = i + 1;
      var colCoordinates = coordinatesOfCol(colNumber);
      return Rule(
          colCoordinates,
          Operation(isUniq, label: 'uniqueness in col', showLabel: false),
          true);
    });
    rules.addAll(colRules);
    return rules;
  }

  /// Is custom rules cover full kenken?
  /// Returns list of cells not covered by custom rules.
  /// If the list is empty, then kenken fully covered by custom rules.
  /// The programmer can create his own logic for filling the kenken with custom rules.
  /// With this method, he can control the completeness of the custom rules.
  List<RC> get notCoveredByCustomRulesCells {
    // prepare coverage filled with `false` for each cell:
    var rulesCoverage = List.generate(size, (_) => List.filled(size, false));
    // now set in `true` each cell for each rule:
    for (var rule in customRules) {
      for (var cellCoordinale in rule.cellsCoordinates) {
        // print('cell ${cellCoordinale.row-1} ${cellCoordinale.col-1} set true');
        rulesCoverage[cellCoordinale.row - 1][cellCoordinale.col - 1] = true;
      }
    }
    // if all cells in `rulesCoverage` are `true`,
    // then kenken full-covered by rules
    var notCoveredCells = <RC>[];
    for (var rowInd = 0; rowInd < rulesCoverage.length; rowInd++) {
      for (var colInd = 0; colInd < rulesCoverage.length; colInd++) {
        // print('$rowInd $colInd ${rulesCoverage[rowInd][colInd]}');
        if (rulesCoverage[rowInd][colInd] == false) {
          notCoveredCells.add(RC(rowInd + 1, colInd + 1));
        }
      }
    }
    return notCoveredCells;
  }

  /// Use it for set cell `value` by cell coordinates `x` and `y`.
  /// Coordinates should be in `1...kenken.size` diapazone.
  void setValue(int row, int col, int value) {
    var errMsg = _coordErrorMessage(row, col);
    if (errMsg.isNotEmpty) {
      throw ArgumentError(errMsg);
    }
    (cells[row - 1])[col - 1].value = value;
  }

  /// Returns list of coordinates for row with given `rowNumber`
  List<RC> coordinatesOfRow(int rowNumber) =>
      List.generate(size, (i) => RC(rowNumber, i + 1));

  /// Returns list of coordinates for column with given `colNumber`
  List<RC> coordinatesOfCol(int colNumber) =>
      List.generate(size, (i) => RC(i + 1, colNumber));

  /// Returns list of coordinates from `list` of `list (size=2)` of `int`.
  /// First digit is row, second digit is column.
  ///
  /// Sample: `cellsOf([[2,1],[2,2],[3,2]])`.
  /// ```
  ///       col1 col2 col3
  ///      -------------------
  /// row1 |     |     |     |
  ///      ------|------------
  /// row2 |  x  |  x  |     |
  ///      ------|------------
  /// row3 |     |  x  |     |
  ///      -------------------
  /// ```
  List<RC> cellsOf(List<List<int>> coordinates) {
    var result = <RC>[];
    for (var rowColumn in coordinates) {
      if (rowColumn.length != 2) {
        throw ArgumentError(
            'each [row,col] should have length=2. But: ${rowColumn}');
      }
      if (rowColumn
          .where((rowOrColumn) => rowOrColumn < 1 || rowOrColumn > size)
          .isNotEmpty) {
        throw ArgumentError('Bad row or column: ${rowColumn}');
      }
      result.add(RC(rowColumn[0], rowColumn[1]));
    }
    return result;
  }

  String _coordErrorMessage(int x, int y) {
    if (x < 1 || x > size || y < 1 || y > size) {
      return 'Bad coordinates: $x,$y';
    }
    return '';
  }

  /// Returns cell value by it coordinates
  int valueOf(RC cellCoord) => cellOf(cellCoord).value;

  /// Return Cell object by it  coordinates
  Cell cellOf(RC cellCoord) {
    var errMsg = _coordErrorMessage(cellCoord.row, cellCoord.col);
    if (errMsg.isNotEmpty) {
      throw ArgumentError(errMsg);
    }
    return cells[cellCoord.row - 1][cellCoord.col - 1];
  }

  /// Returns broken rules for kenken.
  /// If brokenRules list is empty, then kenken are solved.
  /// Any rule is valid (not broken)
  /// if actual result after applying rule predicate to cells values is equal
  /// valid operation result, defined in rule.
  List<Rule> get brokenRules => rules.where((rule) {
        var ruleCellsValues =
            rule.cellsCoordinates.map((coords) => valueOf(coords)).toList();
        var actualResult = rule.operation.predicate(ruleCellsValues);
        return actualResult != rule.validOperationResult;
      }).toList();

  @override
  String toString() {
    // return '$cells';
    var result = [];
    for (var r in rowOrColNumbers) {
      for (var c in rowOrColNumbers) {
        var cell = cellOf(RC(r, c));
        result.add((c == 1 ? '\n' : '') + '($r,$c) $cell');
      }
    }
    return '$result';
  }
}

/// A rule validates list of cells.
/// List of rules validates a whole kenken.
class Rule<ShouldHaveResult> {
  /// `cellsCoordinates` - list of cells that the rule applies to.
  /// `operation` - what to do with cell in the rule. td: список операций
  /// `validOperationResult` - which result is correct.
  Rule(this.cellsCoordinates, this.operation, this.validOperationResult);

  List<RC> cellsCoordinates;
  Operation operation;
  ShouldHaveResult validOperationResult;
  @override
  String toString() =>
      '$operation on $cellsCoordinates should be $validOperationResult';

  /// Answers the question: is the rule has cell with such coordinates.
  /// ```
  /// rule.hasRC(RC(2, 3));
  /// ```
  bool hasRC(RC cell) {
    return cellsCoordinates
        .where((c) => c.row == cell.row && c.col == cell.col)
        .isNotEmpty;
  }

  /// In kenken, the operation and the expected result are displayed
  /// in the topmost and leftmost cell of each group of cells.
  /// This method returns the coordinates of such a cell.
  RC get coordinatesToDisplayTheRule {
    var topmostCell = cellsCoordinates.reduce((RC prevCell, RC currentCell) =>
        prevCell.row < currentCell.row ? prevCell : currentCell);
    var leftmostCell = cellsCoordinates
        .where((cell) => cell.row == topmostCell.row)
        .reduce((prevCell, currentCell) =>
            prevCell.col < currentCell.col ? prevCell : currentCell);
    return leftmostCell;
  }
}

/// RC coordinates. Coordinates should be in 1...kenken.size diapazone.
class RC {
  RC(this.row, this.col);
  int row;
  int col;
  @override
  String toString() => '($row,$col)';
}

class Cell {
  Rule itRule;
  bool isTopBorderBold;
  bool isLeftBorderBold;
  bool isRightBorderBold;
  bool isBottomBorderBold;
  bool doesDisplayRule = false;
  int value;

  /// Returns rule text (if rule text should display in this cell).
  String get ruleText {
    if (!doesDisplayRule) {
      return '';
    }
    var operLblText = itRule.operation.showLabel ? itRule.operation.label : '';
    return '${itRule.validOperationResult}$operLblText';
  }

  @override
  String toString() {
    if (itRule == null) return 'HAS_NOT_RULE_REFERENCE';
    if (doesDisplayRule == null) return 'DOES_DISPLAY_RULE_IS_NULL';
    return 'Txt:[$ruleText] Val:$value L:$isLeftBorderBold R:$isRightBorderBold T:$isTopBorderBold B:$isBottomBorderBold';
  }
}

/// Operation will be apply to list of cells for validate a rule.
/// Operation constructor has `predicate` as positional parameter
/// ( it is function to be apply to cells)
/// `label` - printable operation symbol, accesible by `toString()` method
/// (for display it on kenken or for debug purposes).
/// `showLabel` - whether to show label while kenken rendering.
/// ```
/// Operation(sum, label: '+')
/// ```
/// td: в доке сослаться на список функций
class Operation {
  Operation(this.predicate, {this.label = '', this.showLabel = true});

  String label;
  Predicate predicate;
  bool showLabel;

  @override
  String toString() => label;
}

/// Function which can be applyed to List of values
typedef Predicate = Function(List<int>);

/// Is list of values unique?
bool isUniq(List<int> values) => values.isUniq;

/// Returns single value (first element value)
int singleValue(List<int> values) => values[0];

/// Returns sum of elements
int sum(List<int> values) =>
    values.where((e) => e != null).fold(0, (v, e) => v + e);

/// Returns difference of two values (biggest - smallest)
int diff(List<int> values) {
  if (values[0] == null || values[1] == null) {
    return 0;
  }
  if (values[0] > values[1]) {
    return values[0] - values[1];
  }
  return values[1] - values[0];
}

extension ListUniq<T> on List<T> {
  /// Is List values uniq?
  bool get isUniq {
    if (length < 2) return true;
    for (var i = 0; i < length - 1; i++) {
      if (sublist(i + 1).contains(this[i])) return false;
    }
    return true;
  }
}
